/*------------------------------------------------------------
RSA
  RSA  um sistema de criptografia de chave assimtrica criada 
  em 1977.
   um dos algoritmos mais utilizados na internet para comrcio 
  eletrnico.
  
Funcionamento
  1) O sistema utiliza nmeros primos gigantes na gerao de 
     suas chaves, aproveitando-se do problema de fatorao de 
     grandes inteiros.
  2) Seu algoritmo  relativamente simples de implementar, 
     porm, devida a matemtica pesada envolvida,  de lento 
     funcionamento.

Gerao da chave:
    1) Escolha dois numeros primos bem grandes, "p" e "q";
    2) Faa n = pq;
    3) Faa m = (p-1)(q-1);
    4) Escolha um numero pequeno "e" coprimo de "m", 
       assim: gcd(e, m) = 1;
    5) Encontre "d", tal que: d*e % m = 1 
       assim: d = (1+xm)/e 
       ou em mdulo inverso: d = e^-1 % m
    6) Publique "e" e "n" como a chave publica.
    7) Deixe "d" e "n" como a chave privada.

Codificao:
    C = P^e % n  

Decodificao:
    P = C^d % n  

Referencias:
    http://pajhome.org.uk/crypt/rsa/rsa.html
--------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>	       
#include <math.h>
#include "RSA.h"

#define Int64  long long int
#define Int32  int
#define UInt64 unsigned Int64
#define UInt32 unsigned Int32

//#define DEBUG

//-------------------------------------------------------
// MATEMATICA
//-------------------------------------------------------

// retorna o mximo divisor comum entre a e b
UInt64 gcd(UInt64 a, UInt64 b)
{
	UInt64 c;

	if(a > 1llu && b > 1llu) {
		while (a != b) {
			if(a > b) {
				if(b > 0llu) c = a / b;
				else c = 1llu;
				c *= b;
				if(c >= a) a = b;
				else a -= c;
			} else {
				if(a > 0llu) c = b / a;
				else c = 1llu;
				c *= a;
				if(c >= b) b = a;
				else b -= c;
			}
		}
	} else a = 1llu;

	return a;
}

// testa se o nmero p  primo
int isPrime(UInt32 p) 
{
    if(p == 2) return 1;
    else if(!(p & 1)) return 0;
    else {
         UInt32 n, m = (UInt32) sqrt(p);
         for(n = 3; n < m; n+=2) 
               if(!(p % n)) return 0;
    }
    return 1;
}

// gera um numero primo aleatrio
UInt32 randomPrime()
{
    UInt32 p;
    do {
        do {
            p = rand() | rand()<<16 | 1;                // estatisticamente, 10% dos nmeros so primos
        } while(p <= 0xffffff);
    } while(!isPrime(p));
	return p;
}

// multiplicao modular (a * b mod m)
UInt64 mulMod(UInt64 a, UInt64 b, UInt64 m)
{
    UInt64 r = 0llu, t = 1llu;
    
    t <<= ((sizeof(UInt64)*8)-1);
    
    while(t) {
          r <<= 1;
          if(a & t) r += b;
          r %= m;
          t >>= 1;
    }
      
    return r;
}

// exponenciao modular (x^y mod m)
// referencia: 
// http://www.math.princeton.edu/matalive/Crypto/CryptoLab2/FastPower.html
UInt64 powMod(UInt64 x, UInt64 y, UInt64 m)
{
    UInt64 e = y;
    UInt64 b = x, r = 1llu;
    
    while(e) {
       if(e & 1llu) r = mulMod(r, b, m);   // r = (r*b)%m
       b = mulMod(b, b, m);      //b *= b;
       e >>= 1;
    }
    
    #ifdef DEBUG
           printf("powMod( %I64d, %I64d, %I64d ) = %I64d\n", x, y, m, r);
    #endif
    
    return r;
}

// mdulo do inverso de x
// referencia: 
// http://www.mathreference.com/num-mod,minv.html
// http://en.wikipedia.org/wiki/Extended_Euclidean_algorithm
UInt64 inverseMod(UInt64 a, UInt64 b)
{
    Int64 u[] = {1, 0, a};
    Int64 v[] = {0, 1, b};
    Int64 t[] = {0, 0, 0};
    Int64 q;
    int i;
    
    while(v[2]) {
        q = (u[2] / v[2]);
        for(i = 0; i < 3; i++) {
              t[i] = u[i] - (v[i]*q);
              u[i] = v[i];
              v[i] = t[i];
        }
    }
    while(u[1] < 0) u[1] += a;
    return u[1];
}

//-------------------------------------------------------
// RSA
//-------------------------------------------------------

// procura por um "e" coprimo de "m"
UInt64 forcebruteE(UInt64 m)
{
	UInt64 e = 3llu;                  
    while(gcd(e, m) > 1llu)
    	e += 2llu;
   	return e;
}

// procura por um "d" onde "d*e % m = 1"
UInt64 forcebruteD(UInt64 m, UInt64 e)
{
	UInt64 d, k = 0llu;                  
    while((d=(1+k*m)) % e) k++;
    d /= e;
    return d;
}
       
// gera uma chave RSA publica e privada
void RSA_createKey(char *publicKey, char *privateKey)
{
	UInt64 p, q;
	UInt64 n, m, e, d;		// 64 bits
	UInt64 *key;

    do {
        p = randomPrime();
        q = randomPrime();
        while(p == q) 
          q = randomPrime();       // impede p igual a q
       
        n = p * q;
        m = (p - 1llu) * (q - 1llu);
    
        #ifdef DEBUG
               printf("p = %I64d, q = %I64d, n = %I64d, m = %I64d\n", p, q, n, m);
        #endif
        
        e = forcebruteE(m);
        //d = forcebruteD(m, e);
        d = inverseMod(m, e);
        
    } while(d < 0x0100000000000000llu);
    
    #ifdef DEBUG
           printf("d = %I64d, e = %I64d, de = %I64d\n", d, e, d * e);
    #endif
    
    key = (UInt64 *) publicKey; 
	*key = e; key++;
	*key = n;

    key = (UInt64 *) privateKey; 
	*key = d; key++;
	*key = n;
}

// cifra um bloco de bytes
UInt64 RSA_encrypt(UInt64 P, char * publicKey)
{
	UInt64 *e = (UInt64 *) publicKey;
	UInt64 *n = (e+1);
	return powMod(P, *e, *n);
}

// decifra um bloco de bytes
UInt64 RSA_decrypt(UInt64 C, char * privateKey)
{
	UInt64 *d = (UInt64 *) privateKey;
	UInt64 *n = (d+1);
	return powMod(C, *d, *n);
}

//-------------------------------------------------------------------
// funes principais
//-------------------------------------------------------------------

int rsa_codificar(char *keystr, char *text, char *cript)
{
	int j, size;
	UInt64 *icript = (UInt64 *) cript;
	UInt64 *itext;
	
	size = strlen(text);
    
    for (j=0; j < size; j += sizeof(UInt64)-1) {
	    itext = (UInt64 *) text;
		*icript = RSA_encrypt(*itext & 0x00FFFFFFFFFFFFFFllu, keystr);
	    text += sizeof(UInt64)-1;
	    icript ++;
	}
	*icript = 0;
	
	return -1;
}

int rsa_decodificar(char *keystr, char *cript, char *text, int size)
{
	int j;
	UInt64 *icript = (UInt64 *) cript;
	UInt64 *itext;
    	
    for (j=0; j < size; j += sizeof(UInt64)-1) {
	    itext = (UInt64 *) text;
		*itext = RSA_decrypt(*icript, keystr);
	    text += sizeof(UInt64)-1;
	    icript ++;
	}
	*text = 0;
	
	return -1;
}

/*

// expoente mximo para uma potenciao sobre "base"
#define MAX_INT64 (~((UInt64) 0))
UInt64 maxPow(UInt64 base)
{
   double max = (double) MAX_INT64;
   double expoent = log(max)/log(base);
   return (UInt64) expoent;
}
*/

