/*------------------------------------------------------------
SHA1
  A famlia de SHA (Secure Hash Algorithm) est relacionada com 
  as funes criptogrficas. A funo mais usada nesta famlia, 
  a SHA-1,  usada numa grande variedade de aplicaes e protocolos 
  de segurana, incluindo SSL, SSH e IPSec. SHA-1 foi considerado o 
  sucessor do MD5. Ambos tm vulnerabilidades comprovadas. 
  Em algumas correntes,  sugerido que o SHA-256 ou superior seja 
  usado para tecnologia crtica. 
  Os algoritmos SHA foram desenhados pela National Security Agency 
  (NSA) e publicados como um standard do governo Norte-Americano.

Funcionamento
  Todas as variveis so inteiros de 32 bits sem sinal,
  truncados ao modulo 2^32 durante os clculos.

Referencias:
  http://pt.wikipedia.org/wiki/SHA1
  http://en.wikipedia.org/wiki/SHA1
  http://en.wikipedia.org/wiki/Hmac  
--------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include "sha1.h"

#define UCHAR unsigned char
#define UINT64 unsigned long long
#define UINT32 unsigned int
#define UCHAR unsigned char

// adiciona com mdulo
static UINT32 addmod(UINT32 a, UINT32 b) {
   UINT64 c = a;
   c += b;
   return c & 0xFFFFFFFF; //% 0x100000000ll;
}
       
// faz a rotao  esquerda sem perda de bits
static UINT32 rol(UINT32 num, UINT32 cnt) {
  return (num << cnt) | (num >> (32 - cnt));
}

// copia um inteiro bigendian para um buffer
void cpyIntBE(UCHAR *dest, UINT32 n) {
     int i;
     dest += 3;
     for(i = 0; i < 4; i++) {
        *dest = n & 0xFF;
        dest--; n >>= 8;
     }
}

// l um inteiro bigendian de um buffer
UINT32 readIntBE(UCHAR *src) {
     return (src[0]<<24)|(src[1]<<16)|(src[2]<<8)|(src[3]);
}

// hash de integridade da mensagem
void hash_sha1(UCHAR *message, int sizeOld, UCHAR *hash) {
     UCHAR filler[128], *buffer[2], *buf2;
     UINT32 w[80];
     UINT32 h0, h1, h2, h3, h4;
     UINT32 a, b, c, d, e, f, k, temp;
     UINT32 sizeMsg = sizeOld, sizeFil, size[2];
     UINT32 i, j, m;

     // inicializa as variaveis:
     h0 = 0x67452301;
     h1 = 0xEFCDAB89;
     h2 = 0x98BADCFE;
     h3 = 0x10325476;
     h4 = 0xC3D2E1F0;

     // termina a mensagem original com o caracter 0x80
     // concatena 0s at que esta fique divisvel em blocos de 512 bits
     // sendo que os ultimos 64 bits devem conter o tamanho da mensagem (inteiro bigendian)
     sizeFil = sizeMsg % 64;
     sizeMsg -= sizeFil;
     memset(filler, 0, sizeof(filler));
     memcpy(filler, message+sizeMsg, sizeFil);
     filler[sizeFil] = 0x80;  // terminador da mensagem
     if(sizeFil < 56) {
        sizeFil = sizeof(filler) / 2;
     } else {
        sizeFil = sizeof(filler);
     }
     cpyIntBE(filler+sizeFil-4, sizeOld*8); // grava o tamanho da mensagem (em bits) no final do ltimo bloco
     
     buffer[0] = message;
     buffer[1] = filler;
     size[0] = sizeMsg;
     size[1] = sizeFil;
     
     for(m = 0; m < 2; m++) {
           
         //processa a mensagem em blocos de 512 bits
         for(j = 0; j < size[m]; j+=64) {
    
               // copia os 512 bits do bloco corrente
               // dentro das 16 primeiras posies do
               // vetor de inteiros 32 bits (bigendian)
               buf2 = &buffer[m][j];
               for(i=0; i < 16; i++) {
                    w[i] = readIntBE(buf2);
                    buf2 += 4;
               }
    
               // preenche as demais posies do vetor:
               for(i=16; i < 80; i++) {
                    w[i] = rol(w[i-3] ^ w[i-8] ^ w[i-14] ^ w[i-16], 1);
               }
    
               // inicializa os valores hash do bloco
               a = h0;
               b = h1;
               c = h2;
               d = h3;
               e = h4;
    
               // processa o vetor
               for(i = 0; i < 80; i++) {
                     if(i < 20) {
                          f = d ^ (b & (c ^ d));
                          k = 0x5A827999;
                     } else if(i < 40) {
                          f = b ^ c ^ d;
                          k = 0x6ED9EBA1;
                     } else if(i < 60) {
                          f = (b & c) | (d & (b | c));
                          k = 0x8F1BBCDC;
                     } else {
                          f = b ^ c ^ d;
                          k = 0xCA62C1D6;
                     }
    
                     temp = rol(a, 5);
                     temp = addmod(temp, f);
                     temp = addmod(temp, e);
                     temp = addmod(temp, k);
                     temp = addmod(temp, w[i]);
                     
                     e = d;
                     d = c;
                     c = rol(b, 30);
                     b = a;
                     a = temp;
               }
            
               // adiciona o bloco ao hash global
               h0 = addmod(h0, a);
               h1 = addmod(h1, b);
               h2 = addmod(h2, c);
               h3 = addmod(h3, d);
               h4 = addmod(h4, e);
         }
         
     }
     
     // produz o hash final de 160 bits (bigendian), 
     // hash = h0 concat h1 concat h2 concat h3 concat h4
     cpyIntBE(hash, h0); hash+=4;
     cpyIntBE(hash, h1); hash+=4;
     cpyIntBE(hash, h2); hash+=4;
     cpyIntBE(hash, h3); hash+=4;
     cpyIntBE(hash, h4); hash+=4;
        
}

// testa o algoritmo
int sha1_test() {
    UCHAR hash[20];
    UCHAR test[] = "abc";
    UCHAR result[] = {0xa9, 0x99, 0x3e, 0x36, 0x47, 0x06, 0x81, 0x6a, 0xba, 0x3e, 0x25, 0x71, 0x78, 0x50, 0xc2, 0x6c, 0x9c, 0xd0, 0xd8, 0x9d};
    hash_sha1(test, sizeof(test), hash);
    return strncmp((char *) result, (char *) hash, sizeof(hash));
}

// hash de autenticidade da mensagem
void hmac_sha1(UCHAR *key, int sizeKey, UCHAR *message, int sizeMsg, UCHAR *hash) {
     UCHAR hkey[20];
     UCHAR opad[sizeof(hkey)], ipad[sizeof(hkey)];
     UCHAR *buf;
     int i;
     
     if(sizeMsg < 0) {
        memset(hash, 0, sizeof(hkey));
        return;
     }
     
     memset(opad, 0x5c, sizeof(opad));
     memset(ipad, 0x5c, sizeof(opad));
     
     if(sizeKey > sizeof(hkey)) {
        hash_sha1(message, sizeMsg, hkey);
        sizeKey = sizeof(hkey);
     } else {
        memset(hkey, 0, sizeof(hkey));
        memcpy(hkey, key, sizeKey);
     }
       
     for(i=0; i < sizeKey; i++) {
        ipad[i] ^= hkey[i];
        opad[i] ^= hkey[i];
     }

     // concatena ipad com a mensagem
     buf = (UCHAR *) malloc(sizeof(ipad)+sizeMsg+100);
     if(buf) {
         memcpy(buf, ipad, sizeof(ipad));
         memcpy(buf+sizeof(ipad), message, sizeMsg);
         
         // calcula novo hash da chave
         hash_sha1(buf, sizeof(ipad)+sizeMsg, hkey);
         
         // concatena opad com o novo hash
         memcpy(buf, opad, sizeof(opad));
         memcpy(buf+sizeof(opad), hkey, sizeof(hkey));
         
         // calcula o hash definitivo
         hash_sha1(buf, sizeof(opad)+sizeof(hkey), hash);
         
         free(buf);
     }
}
 
//--------------------------------------------------------------
    
void sha1_integridade(char *message, char *hash) {
     hash_sha1((UCHAR *) message, strlen(message), (UCHAR *) hash);
}

void sha1_autenticidade(char *key, char *message, char *hash) {
     hmac_sha1((UCHAR *) key, strlen(key), (UCHAR *) message, strlen(message), (UCHAR *) hash);
}

